/*******************************************************************************
 * Copyright (c) 2000, 2011 IBM Corporation and others.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 *     IBM Corporation - initial API and implementation
 *     Syntevo GmbH    - major redesign
 *******************************************************************************/
package com.syntevo.q.swt;

import java.util.*;

import org.eclipse.swt.*;
import org.eclipse.swt.graphics.*;
import org.eclipse.swt.widgets.*;

final class QTabFolderRenderer {

	// Constants ==============================================================

	private static final String COLOR_CACHE = "syntevo.TFR.ColorCache"; //$NON-NLS-1$

	private static final int[] SIMPLE_TOP_LEFT_CORNER = new int[] {0, 2, 1, 1, 2, 0};
	private static final int[] SIMPLE_TOP_RIGHT_CORNER = new int[] {-2, 0, -1, 1, 0, 2};
	private static final int[] SIMPLE_UNSELECTED_INNER_CORNER = new int[] {0, 0};

	private static final int[] SIMPLE_TOP_RIGHT_CORNER_BORDERLESS = new int[] {-3, 0, -2, 1, -1, 2};

	private static final RGB CLOSE_FILL = new RGB(252, 160, 160);

	private static final int BUTTON_SIZE = 18;

	private static final int BUTTON_BORDER = SWT.COLOR_WIDGET_DARK_SHADOW;
	private static final int BUTTON_FILL = SWT.COLOR_LIST_BACKGROUND;
	private static final int BORDER1_COLOR = SWT.COLOR_WIDGET_NORMAL_SHADOW;

	private static final int ITEM_TOP_MARGIN = 2;
	private static final int ITEM_BOTTOM_MARGIN = 2;
	private static final int ITEM_LEFT_MARGIN = 4;
	private static final int ITEM_RIGHT_MARGIN = 4;
	public static final int INTERNAL_SPACING = 4;
	private static final int FLAGS = SWT.DRAW_TRANSPARENT | SWT.DRAW_MNEMONIC;
	private static final String ELLIPSIS = "..."; //$NON-NLS-1$

	//Part constants
	public static final int PART_BODY = -1;
	public static final int PART_HEADER = -2;
	public static final int PART_BORDER = -3;
	public static final int PART_CHEVRON_BUTTON = -7;
	public static final int PART_CLOSE_BUTTON = -8;

	public static final int MINIMUM_SIZE = 1 << 24; //TODO: Should this be a state?
	public static final int MIN_CHARS = 20;

	// Fields =================================================================

	private final QTabFolder tabFolder;

	private Color fillColor;

	// Setup ==================================================================

	public QTabFolderRenderer(QTabFolder tabFolder) {
		if (tabFolder == null) {
			SWT.error(SWT.ERROR_NULL_ARGUMENT);
		}
		this.tabFolder = tabFolder;
	}

	// Accessing ==============================================================

	public int getMinTabWidth(int i, GC gc) {
		return getTabSize(i, gc, i == tabFolder.getSelectedIndex(), true).x;
	}

	public int getTabWidth(int i, GC gc) {
		return getTabSize(i, gc, i == tabFolder.getSelectedIndex(), false).x;
	}

	public Point getChevronButtonSize() {
		return new Point(3 * BUTTON_SIZE / 2, BUTTON_SIZE);
	}

	public Point getCloseButtonSize() {
		return new Point(BUTTON_SIZE, BUTTON_SIZE);
	}

	public int getHeaderHeight() {
		final int fixedTabHeight = tabFolder.getFixedTabHeight();
		if (fixedTabHeight != SWT.DEFAULT) {
			return fixedTabHeight + 1;
		}

		final GC gc = new GC(tabFolder);
		final int itemCount = tabFolder.getItemCount();
		int height = 0;
		if (itemCount == 0) {
			height = gc.textExtent("Default", FLAGS).y + ITEM_TOP_MARGIN + ITEM_BOTTOM_MARGIN; //$NON-NLS-1$
		}
		else {
			for (int i = 0; i < itemCount; i++) {
				height = Math.max(height, getTabSize(i, gc, false, false).y);
			}
		}
		gc.dispose();
		return height;
	}

	public Rectangle computeBodyTrim(int x, int y, int width, int height) {
		final int borderSize = tabFolder.isBorderVisible() ? 1 : 0;
		final int tabHeight = tabFolder.getTabHeight();
		final int style = tabFolder.getStyle();
		int highlight_header = (style & SWT.FLAT) != 0 ? 1 : 3;
		final int highlight_margin = (style & SWT.FLAT) != 0 ? 0 : 2;
		if (tabFolder.getFixedTabHeight() == 0 && (style & SWT.FLAT) != 0 && (style & SWT.BORDER) == 0) {
			highlight_header = 0;
		}
		x = x - highlight_margin - borderSize;
		width = width + 2 * borderSize + 2 * highlight_margin;
		y = y - highlight_header - tabHeight;
		height = height + borderSize + tabHeight + highlight_header + highlight_margin;
		return new Rectangle(x, y, width, height);
	}

	public Rectangle computeBorderTrim() {
		final int borderSize = tabFolder.isBorderVisible() ? 1 : 0;
		return new Rectangle(-borderSize, 0, 2 * borderSize, borderSize);
	}

	public Rectangle computeHeaderTrim(int x, int y, int width, int height) {
		return new Rectangle(x, y, width, height);
	}

	public Rectangle computeTabTrim(int width, int height) {
		width += ITEM_LEFT_MARGIN + ITEM_RIGHT_MARGIN;
		height += ITEM_TOP_MARGIN + ITEM_BOTTOM_MARGIN;
		return new Rectangle(-ITEM_LEFT_MARGIN, -ITEM_TOP_MARGIN, width, height);
	}

	/**
	 * Dispose of any operating system resources associated with
	 * the renderer. Called by the CTabFolder parent upon receiving
	 * the dispose event or when changing the renderer.
	 *
	 * @since 3.6
	 */
	public void dispose() {
		if (fillColor != null) {
			releaseColor(tabFolder.getDisplay(), fillColor);
			fillColor = null;
		}
	}

	public void drawTab(int part, int state, Rectangle bounds, GC gc) {
		if (part < 0) {
			SWT.error(SWT.ERROR_UNSPECIFIED);
		}
		if (part >= tabFolder.getItemCount()) {
			SWT.error(SWT.ERROR_UNSPECIFIED);
		}

		if (bounds.width == 0 || bounds.height == 0) {
			return;
		}

		if ((state & SWT.SELECTED) != 0) {
			drawSelected(part, gc, bounds, (state & SWT.BACKGROUND) != 0, (state & SWT.FOREGROUND) != 0);
		}
		else {
			drawUnselected(part, gc, bounds, (state & SWT.BACKGROUND) != 0, (state & SWT.FOREGROUND) != 0);
		}
	}

	public void drawBody(Point size, GC gc) {
		if (size == null) {
			SWT.error(SWT.ERROR_NULL_ARGUMENT);
		}
		if (gc == null) {
			SWT.error(SWT.ERROR_NULL_ARGUMENT);
		}

		final int selectedIndex = tabFolder.getSelectedIndex();
		final int tabHeight = tabFolder.getTabHeight();

		final int borderLeft = tabFolder.isBorderVisible() ? 1 : 0;

		final int style = tabFolder.getStyle();
		final int highlight_header = (style & SWT.FLAT) != 0 ? 1 : 3;
		final int highlight_margin = (style & SWT.FLAT) != 0 ? 0 : 2;

		// fill in body
		final int width = size.x - borderLeft - borderLeft - 2 * highlight_margin;
		final int height = size.y - borderLeft - tabHeight - highlight_header - highlight_margin;
		// Draw highlight margin
		if (highlight_margin > 0) {
			final int y1 = tabHeight + highlight_header;
			final int x2 = size.x - borderLeft;
			final int y2 = size.y - borderLeft;
			final int[] shape = new int[] {borderLeft, y1,
					borderLeft + highlight_margin, y1,
					borderLeft + highlight_margin, y2 - highlight_margin,
					x2 - highlight_margin, y2 - highlight_margin,
					x2 - highlight_margin, y1,
					x2, y1,
					x2, y2,
					borderLeft, y2};
			// If horizontal gradient, show gradient across the whole area
			gc.setBackground(selectedIndex == -1 ? tabFolder.getBackground() : tabFolder.getSelectionBackground());
			gc.fillPolygon(shape);
		}
		//Draw client area
		if ((tabFolder.getStyle() & SWT.NO_BACKGROUND) != 0) {
			gc.setBackground(tabFolder.getBackground());
			final int xClient = borderLeft + highlight_margin;
			final int yClient;
			yClient = tabHeight + highlight_header;
			gc.fillRectangle(xClient, yClient, width, height);
		}

		//draw 1 pixel border left, below and right to the content
		if (borderLeft > 0) {
			gc.setForeground(tabFolder.getDisplay().getSystemColor(BORDER1_COLOR));
			final int x1 = borderLeft - 1;
			final int x2 = size.x - borderLeft;
			final int y2 = size.y - borderLeft;
			gc.drawLine(x1, tabHeight, x1, y2); // left
			gc.drawLine(x2, tabHeight, x2, y2); // right
			gc.drawLine(x1, y2, x2, y2); // bottom
		}
	}

	public void drawChevron(int chevronImageState, Rectangle chevronRect, GC gc) {
		if (chevronRect.width == 0 || chevronRect.height == 0) {
			return;
		}

		// draw chevron (10x7)
		final Display display = tabFolder.getDisplay();
		final Point dpi = display.getDPI();
		final int fontHeight = 72 * 10 / dpi.y;
		final FontData fd = tabFolder.getFont().getFontData()[0];
		fd.setHeight(fontHeight);
		final Font f = new Font(display, fd);
		final int fHeight = f.getFontData()[0].getHeight() * dpi.y / 72;
		final int indent = Math.max(2, (chevronRect.height - fHeight - 4) / 2);
		final int x = chevronRect.x + 2;
		final int y = chevronRect.y + indent;
		final int count;
		final int itemCount = tabFolder.getItemCount();
		int showCount = 0;
		while (showCount < tabFolder.getPriority().length && tabFolder.getItem(tabFolder.getPriority()[showCount]).isShowing()) {
			showCount++;
		}
		count = itemCount - showCount;
		final String chevronString = count > 99 ? "99+" : String.valueOf(count); //$NON-NLS-1$
		switch (chevronImageState & (SWT.HOT | SWT.SELECTED)) {
		case SWT.NONE: {
			final Color chevronBorder = tabFolder.getForeground();
			gc.setForeground(chevronBorder);
			gc.setFont(f);
			gc.drawLine(x, y, x + 2, y + 2);
			gc.drawLine(x + 2, y + 2, x, y + 4);
			gc.drawLine(x + 1, y, x + 3, y + 2);
			gc.drawLine(x + 3, y + 2, x + 1, y + 4);
			gc.drawLine(x + 4, y, x + 6, y + 2);
			gc.drawLine(x + 6, y + 2, x + 5, y + 4);
			gc.drawLine(x + 5, y, x + 7, y + 2);
			gc.drawLine(x + 7, y + 2, x + 4, y + 4);
			gc.drawString(chevronString, x + 7, y + 3, true);
			break;
		}
		case SWT.HOT: {
			gc.setForeground(display.getSystemColor(BUTTON_BORDER));
			gc.setBackground(display.getSystemColor(BUTTON_FILL));
			gc.setFont(f);
			gc.fillRoundRectangle(chevronRect.x, chevronRect.y, chevronRect.width, chevronRect.height, 6, 6);
			gc.drawRoundRectangle(chevronRect.x, chevronRect.y, chevronRect.width - 1, chevronRect.height - 1, 6, 6);
			gc.drawLine(x, y, x + 2, y + 2);
			gc.drawLine(x + 2, y + 2, x, y + 4);
			gc.drawLine(x + 1, y, x + 3, y + 2);
			gc.drawLine(x + 3, y + 2, x + 1, y + 4);
			gc.drawLine(x + 4, y, x + 6, y + 2);
			gc.drawLine(x + 6, y + 2, x + 5, y + 4);
			gc.drawLine(x + 5, y, x + 7, y + 2);
			gc.drawLine(x + 7, y + 2, x + 4, y + 4);
			gc.drawString(chevronString, x + 7, y + 3, true);
			break;
		}
		case SWT.SELECTED: {
			gc.setForeground(display.getSystemColor(BUTTON_BORDER));
			gc.setBackground(display.getSystemColor(BUTTON_FILL));
			gc.setFont(f);
			gc.fillRoundRectangle(chevronRect.x, chevronRect.y, chevronRect.width, chevronRect.height, 6, 6);
			gc.drawRoundRectangle(chevronRect.x, chevronRect.y, chevronRect.width - 1, chevronRect.height - 1, 6, 6);
			gc.drawLine(x + 1, y + 1, x + 3, y + 3);
			gc.drawLine(x + 3, y + 3, x + 1, y + 5);
			gc.drawLine(x + 2, y + 1, x + 4, y + 3);
			gc.drawLine(x + 4, y + 3, x + 2, y + 5);
			gc.drawLine(x + 5, y + 1, x + 7, y + 3);
			gc.drawLine(x + 7, y + 3, x + 6, y + 5);
			gc.drawLine(x + 6, y + 1, x + 8, y + 3);
			gc.drawLine(x + 8, y + 3, x + 5, y + 5);
			gc.drawString(chevronString, x + 8, y + 4, true);
			break;
		}
		}
		f.dispose();
	}

	public void drawHeader(GC gc) {
		final Point size = tabFolder.getSize();
		final int tabHeight = tabFolder.getTabHeight();
		final int style = tabFolder.getStyle();

		final int borderLeft = tabFolder.isBorderVisible() ? 1 : 0;

		final int highlight_header = (style & SWT.FLAT) != 0 ? 1 : 3;

		final int x = Math.max(0, borderLeft - 1);
		final int width = size.x - borderLeft - borderLeft + 1;
		final int height = tabHeight - 1;
		// Draw Tab Header
		final int[] left = SIMPLE_TOP_LEFT_CORNER;
		final int[] right;
		if ((style & SWT.BORDER) != 0) {
			right = SIMPLE_TOP_RIGHT_CORNER;
		}
		else {
			right = SIMPLE_TOP_RIGHT_CORNER_BORDERLESS;
		}

		final int[] shape = new int[left.length + right.length + 4];
		int index = 0;
		shape[index++] = x;
		shape[index++] = height + highlight_header + 1;
		for (int i = 0; i < left.length / 2; i++) {
			shape[index++] = x + left[2 * i];
			shape[index++] = left[2 * i + 1];
		}
		for (int i = 0; i < right.length / 2; i++) {
			shape[index++] = x + width + right[2 * i];
			shape[index++] = right[2 * i + 1];
		}
		shape[index++] = x + width;
		shape[index] = height + highlight_header + 1;
		// Fill in background
		drawBackground(gc, shape, false);
		// Fill in parent background for non-rectangular shape
		final Region r = new Region();
		r.add(new Rectangle(x, 0, width + 1, height + 1));
		r.subtract(shape);
		gc.setBackground(tabFolder.getParent().getBackground());
		fillRegion(gc, r);
		r.dispose();

		final Color borderColor = tabFolder.getDisplay().getSystemColor(BORDER1_COLOR);

		// Draw border line (#)
		// +---------+#################################
		// |         |                                #
		// |         |                                #
		if (borderLeft > 0) {
			gc.setForeground(borderColor);
			gc.drawPolyline(shape);
		}
	}

	// Utils ==================================================================

	private Point getTabSize(int tabIndex, GC gc, boolean selected, boolean minimum) {
		if (tabIndex < 0) {
			SWT.error(SWT.ERROR_UNSPECIFIED);
		}
		if (tabIndex > tabFolder.getItemCount()) {
			SWT.error(SWT.ERROR_UNSPECIFIED);
		}

		final QTabFolderItem item = tabFolder.getItem(tabIndex);
		if (item.isDisposed()) {
			return new Point(0, 0);
		}

		int width = 0;
		int height = 0;

		String text;
		if (minimum) {
			text = item.getText();
			if (text != null && text.length() > MIN_CHARS) {
				final int end = MIN_CHARS - ELLIPSIS.length();
				text = text.substring(0, end);
				text += ELLIPSIS;
			}
		}
		else {
			text = item.getText();
		}

		if (text != null) {
			if (width > 0) {
				width += INTERNAL_SPACING;
			}
			final Font font = tabFolder.getFont();
			if (font == null) {
				final Point size = gc.textExtent(text, FLAGS);
				width += size.x;
				height = Math.max(height, size.y);
			}
			else {
				final Font gcFont = gc.getFont();
				gc.setFont(font);
				final Point size = gc.textExtent(text, FLAGS);
				width += size.x;
				height = Math.max(height, size.y);
				gc.setFont(gcFont);
			}
		}

		if (tabFolder.isShowClose() || item.isShowClose()) {
			if (selected || tabFolder.isShowUnselectedClose()) {
				if (width > 0) {
					width += INTERNAL_SPACING;
				}
				width += getCloseButtonSize().x;
			}
		}

		final Rectangle trim = computeTabTrim(width, height);
		return new Point(trim.width, trim.height);
	}

	private void drawBackground(GC gc, int[] shape, boolean selected) {
		final Color defaultBackground = selected ? tabFolder.getSelectionBackground() : tabFolder.getBackground();
		final Point size = tabFolder.getSize();
		int width = size.x;
		final int height = tabFolder.getTabHeight() + ((tabFolder.getStyle() & SWT.FLAT) != 0 ? 1 : 3);
		int x = 0;

		final int borderLeft = tabFolder.isBorderVisible() ? 1 : 0;
		final int borderTop = 0;

		if (borderLeft > 0) {
			x += 1;
			width -= 2;
		}
		drawBackground(gc, shape, x, borderTop, width, height, defaultBackground);
	}

	private void drawBackground(GC gc, int[] shape, int x, int y, int width, int height, Color defaultBackground) {
		Region clipping = null;
		Region region = null;
		if (shape != null) {
			clipping = new Region();
			gc.getClipping(clipping);
			region = new Region();
			region.add(shape);
			region.intersect(clipping);
			gc.setClipping(region);
		}

		if ((tabFolder.getStyle() & SWT.NO_BACKGROUND) != 0 || !defaultBackground.equals(tabFolder.getBackground())) {
			gc.setBackground(defaultBackground);
			gc.fillRectangle(x, y, width, height);
		}

		if (shape != null) {
			gc.setClipping(clipping);
			clipping.dispose();
			region.dispose();
		}
	}

	private void drawBorder(GC gc, int[] shape) {
		gc.setForeground(tabFolder.getDisplay().getSystemColor(BORDER1_COLOR));
		gc.drawPolyline(shape);
	}

	private void drawClose(GC gc, Rectangle closeRect, int closeImageState) {
		if (closeRect.width == 0 || closeRect.height == 0) {
			return;
		}

		final Display display = tabFolder.getDisplay();

		// draw X 9x9
		final int x = closeRect.x + Math.max(1, (closeRect.width - 9) / 2);
		final int y = closeRect.y + Math.max(1, (closeRect.height - 9) / 2) + 1;

		final Color closeBorder = display.getSystemColor(BUTTON_BORDER);

		closeImageState &= SWT.HOT | SWT.SELECTED | SWT.BACKGROUND;
		if (closeImageState == SWT.NONE) {
			final int[] shape = new int[] {x, y, x + 2, y, x + 4, y + 2, x + 5, y + 2, x + 7, y, x + 9, y,
					x + 9, y + 2, x + 7, y + 4, x + 7, y + 5, x + 9, y + 7, x + 9, y + 9,
					x + 7, y + 9, x + 5, y + 7, x + 4, y + 7, x + 2, y + 9, x, y + 9,
					x, y + 7, x + 2, y + 5, x + 2, y + 4, x, y + 2};
			gc.setBackground(display.getSystemColor(BUTTON_FILL));
			gc.fillPolygon(shape);
			gc.setForeground(closeBorder);
			gc.drawPolygon(shape);
		}
		else if (closeImageState == SWT.HOT) {
			final int[] shape = new int[] {x, y, x + 2, y, x + 4, y + 2, x + 5, y + 2, x + 7, y, x + 9, y,
					x + 9, y + 2, x + 7, y + 4, x + 7, y + 5, x + 9, y + 7, x + 9, y + 9,
					x + 7, y + 9, x + 5, y + 7, x + 4, y + 7, x + 2, y + 9, x, y + 9,
					x, y + 7, x + 2, y + 5, x + 2, y + 4, x, y + 2};
			gc.setBackground(getFillColor());
			gc.fillPolygon(shape);
			gc.setForeground(closeBorder);
			gc.drawPolygon(shape);
		}
		else if (closeImageState == SWT.SELECTED) {
			final int[] shape = new int[] {x + 1, y + 1, x + 3, y + 1, x + 5, y + 3, x + 6, y + 3, x + 8, y + 1, x + 10, y + 1,
					x + 10, y + 3, x + 8, y + 5, x + 8, y + 6, x + 10, y + 8, x + 10, y + 10,
					x + 8, y + 10, x + 6, y + 8, x + 5, y + 8, x + 3, y + 10, x + 1, y + 10,
					x + 1, y + 8, x + 3, y + 6, x + 3, y + 5, x + 1, y + 3};
			gc.setBackground(getFillColor());
			gc.fillPolygon(shape);
			gc.setForeground(closeBorder);
			gc.drawPolygon(shape);
		}
		else if (closeImageState == SWT.BACKGROUND) {
			final int[] shape = new int[] {x, y, x + 10, y, x + 10, y + 10, x, y + 10};
			drawBackground(gc, shape, false);
		}
	}

	/*
	 * Draw the unselected border for the receiver on the left.
	 *
	 * @param gc
	 */
	private void drawLeftUnselectedBorder(GC gc, Rectangle bounds) {
		final int x = bounds.x;
		final int y = bounds.y;
		final int height = bounds.height;

		final int[] left = SIMPLE_UNSELECTED_INNER_CORNER;
		final int[] shape = new int[left.length + 2];
		int index = 0;
		shape[index++] = x;
		shape[index++] = y + height;
		for (int i = 0; i < left.length / 2; i++) {
			shape[index++] = x + left[2 * i];
			shape[index++] = y + left[2 * i + 1];
		}

		drawBorder(gc, shape);
	}

	/*
	 * Draw the unselected border for the receiver on the right.
	 *
	 * @param gc
	 */
	private void drawRightUnselectedBorder(GC gc, Rectangle bounds) {
		final int x = bounds.x;
		final int y = bounds.y;
		final int width = bounds.width;
		final int height = bounds.height;

		final int startX = x + width - 1;

		final int[] right = SIMPLE_UNSELECTED_INNER_CORNER;

		final int[] shape = new int[right.length + 2];

		int index = 0;
		for (int i = 0; i < right.length / 2; i++) {
			shape[index++] = startX + right[2 * i];
			shape[index++] = y + right[2 * i + 1];
		}

		shape[index++] = startX;
		shape[index] = y + height;

		drawBorder(gc, shape);
	}

	private void drawSelected(int itemIndex, GC gc, Rectangle bounds, boolean background, boolean foreground) {
		final QTabFolderItem item = tabFolder.getItem(itemIndex);
		final int x = bounds.x;
		final int y = bounds.y;
		final int height = bounds.height;
		final int width = bounds.width;

		final Point size = tabFolder.getSize();

		final int rightEdge = Math.min(x + width, tabFolder.getRightItemEdge());
		//	 Draw selection border across all tabs

		if (background) {
			final int highlight_header = (tabFolder.getStyle() & SWT.FLAT) != 0 ? 1 : 3;
			gc.setBackground(tabFolder.getSelectionBackground());
			gc.fillRectangle(1, tabFolder.getTabHeight() + 1, size.x - 2, highlight_header - 1);

			if (!item.isShowing()) {
				final int x1 = 0;
				final int y1 = y + height;
				final int x2 = size.x - 1;
				gc.setForeground(tabFolder.getDisplay().getSystemColor(BORDER1_COLOR));
				gc.drawLine(x1, y1, x2, y1);
				return;
			}

			// draw selected tab background and outline
			final int[] left = SIMPLE_TOP_LEFT_CORNER;
			final int[] right = SIMPLE_TOP_RIGHT_CORNER;

			final int[] shape = new int[left.length + right.length + 8];
			int index = 0;
			shape[index++] = x; // first point repeated here because below we reuse shape to draw outline
			shape[index++] = y + height + 1;
			shape[index++] = x;
			shape[index++] = y + height + 1;
			for (int i = 0; i < left.length / 2; i++) {
				shape[index++] = x + left[2 * i];
				shape[index++] = y + left[2 * i + 1];
			}
			for (int i = 0; i < right.length / 2; i++) {
				shape[index++] = rightEdge - 1 + right[2 * i];
				shape[index++] = y + right[2 * i + 1];
			}
			shape[index++] = rightEdge - 1;
			shape[index++] = y + height + 1;
			shape[index++] = rightEdge - 1;
			shape[index] = y + height + 1;

			final Rectangle clipping = gc.getClipping();
			final Rectangle clipBounds = item.getBounds();
			clipBounds.height += 1;
			if (clipping.intersects(clipBounds)) {
				// fill in tab background
				final Color defaultBackground = tabFolder.getSelectionBackground();
				drawBackground(gc, shape, x, y + 1, width, height, defaultBackground);
			}

			// draw outline
			shape[0] = 0;
			shape[shape.length - 2] = size.x;
			for (int i = 0; i < shape.length / 2; i++) {
				if (shape[2 * i + 1] == y + height + 1) {
					shape[2 * i + 1] -= 1;
				}
			}

			final Color borderColor = tabFolder.getDisplay().getSystemColor(BORDER1_COLOR);
			gc.setForeground(borderColor);
			gc.drawPolyline(shape);

			if (!clipping.intersects(clipBounds)) {
				return;
			}
		}

		if (foreground) {
			// draw Image
			final Rectangle trim = computeTabTrim(0, 0);
			final int xDraw = x - trim.x;

			// draw Text
			int textWidth = rightEdge - xDraw - (trim.width + trim.x);
			if (item.getCloseRect().width > 0) {
				textWidth -= item.getCloseRect().width + INTERNAL_SPACING;
			}
			if (textWidth > 0) {
				final Font gcFont = gc.getFont();
				gc.setFont(tabFolder.getFont());

				if (item.getShortenedText() == null || item.getShortenedTextWidth() != textWidth) {
					item.setShortenedText(shortenText(gc, item.getText(), textWidth));
					item.setShortenedTextWidth(textWidth);
				}
				final Point extent = gc.textExtent(item.getShortenedText(), FLAGS);
				int textY = y + (height - extent.y) / 2;
				textY += 1;

				gc.setForeground(tabFolder.getSelectionForeground());
				gc.drawText(item.getShortenedText(), xDraw, textY, FLAGS);
				gc.setFont(gcFont);
			}
			if (tabFolder.isShowClose() || item.isShowClose()) {
				drawClose(gc, item.getCloseRect(), item.getCloseImageState());
			}
		}
	}

	private void drawUnselected(int index, GC gc, Rectangle bounds, boolean background, boolean foreground) {
		final QTabFolderItem item = tabFolder.getItem(index);
		final int x = bounds.x;
		final int y = bounds.y;
		final int height = bounds.height;
		final int width = bounds.width;

		// Do not draw partial items
		if (!item.isShowing()) {
			return;
		}

		final Rectangle clipping = gc.getClipping();
		if (!clipping.intersects(bounds)) {
			return;
		}

		if (background) {
			if (index > 0 && index < tabFolder.getSelectedIndex()) {
				drawLeftUnselectedBorder(gc, bounds);
			}
			// If it is the last one then draw a line
			if (index > tabFolder.getSelectedIndex()) {
				drawRightUnselectedBorder(gc, bounds);
			}
		}

		if (foreground) {
			// draw Image
			final Rectangle trim = computeTabTrim(0, 0);
			final int xDraw = x - trim.x;
			// draw Text
			int textWidth = x + width - xDraw - (trim.width + trim.x);
			if (tabFolder.isShowUnselectedClose() && (tabFolder.isShowClose() || item.isShowClose())) {
				textWidth -= item.getCloseRect().width + INTERNAL_SPACING;
			}
			if (textWidth > 0) {
				final Font gcFont = gc.getFont();
				gc.setFont(tabFolder.getFont());
				if (item.getShortenedText() == null || item.getShortenedTextWidth() != textWidth) {
					item.setShortenedText(shortenText(gc, item.getText(), textWidth));
					item.setShortenedTextWidth(textWidth);
				}
				final Point extent = gc.textExtent(item.getShortenedText(), FLAGS);
				int textY = y + (height - extent.y) / 2;
				textY += 1;
				gc.setForeground(tabFolder.getForeground());
				gc.drawText(item.getShortenedText(), xDraw, textY, FLAGS);
				gc.setFont(gcFont);
			}
			// draw close
			if (tabFolder.isShowUnselectedClose() && (tabFolder.isShowClose() || item.isShowClose())) {
				drawClose(gc, item.getCloseRect(), item.getCloseImageState());
			}
		}
	}

	private void fillRegion(GC gc, Region region) {
		// NOTE: region passed in to this function will be modified
		final Region clipping = new Region();
		gc.getClipping(clipping);
		region.intersect(clipping);
		gc.setClipping(region);
		gc.fillRectangle(region.getBounds());
		gc.setClipping(clipping);
		clipping.dispose();
	}

	private Color getFillColor() {
		if (fillColor == null) {
			fillColor = getColor(tabFolder.getDisplay(), CLOSE_FILL);
		}
		return fillColor;
	}

	private String shortenText(GC gc, String text, int width) {
		if (gc.textExtent(text, FLAGS).x <= width) {
			return text;
		}

		final int ellipseWidth = gc.textExtent(ELLIPSIS, FLAGS).x;
		final int length = text.length();
		final TextLayout layout = new TextLayout(tabFolder.getDisplay());
		layout.setText(text);
		int end = layout.getPreviousOffset(length, SWT.MOVEMENT_CLUSTER);
		while (end > 0) {
			text = text.substring(0, end);
			final int l = gc.textExtent(text, FLAGS).x;
			if (l + ellipseWidth <= width) {
				break;
			}
			end = layout.getPreviousOffset(end, SWT.MOVEMENT_CLUSTER);
		}
		layout.dispose();
		return end == 0 ? text.substring(0, 1) : text + ELLIPSIS;
	}

	@SuppressWarnings({"unchecked"})
	private static Color getColor(Display display, RGB rgb) {
		final Color color;
		Map<RGB, Object[]> colorCache = (Map<RGB, Object[]>)display.getData(COLOR_CACHE);
		if (colorCache == null) {
			colorCache = new HashMap<RGB, Object[]>();
			display.setData(COLOR_CACHE, colorCache);
		}

		final Object[] colorData = colorCache.get(rgb);
		if (colorData != null) {
			color = (Color)colorData[0];
			final int refcount = ((Integer)colorData[1]).intValue();
			colorData[1] = new Integer(refcount + 1);
		}
		else {
			color = new Color(display, rgb);
			colorCache.put(rgb, new Object[] {color, new Integer(1)});
		}
		return color;
	}

	@SuppressWarnings({"unchecked"})
	private static void releaseColor(Display display, Color color) {
		final RGB rgb = color.getRGB();
		Map<RGB, Object[]> ColorCache = (Map<RGB, Object[]>)display.getData(COLOR_CACHE);
		if (ColorCache == null) {
			ColorCache = new HashMap<RGB, Object[]>();
			display.setData(COLOR_CACHE, ColorCache);
		}
		final Object[] colorData = ColorCache.get(rgb);
		if (colorData != null) {
			int refcount = ((Integer)colorData[1]).intValue();
			refcount--;
			if (refcount == 0) {
				ColorCache.remove(rgb);
				color.dispose();
			}
			else {
				colorData[1] = new Integer(refcount);
			}
		}
	}
}
